/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *	fssubs.c: fs utility subroutines
 */

#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mntent.h>

#include "jfs_types.h"
#include "jfs_superblock.h"
#include "jfs_filsys.h"
#include "jfs_dinode.h"
#include "jfs_dmap.h"
#include "jfs_imap.h"
#include "jfs_debug.h"

#include "devices.h"
#include "debug.h"

#include "fssubs.h"
#include "utilsubs.h"
#include "message.h"

/* forward references */
static void bAllocate(void);

/* Global Data */

s32	Fail_Quietly = 0;	/* If set, don't report errors */

HFILE	LVHandle;
LV_t 	LVMount;	/* logical volume descriptor */
LV_t	*lvMount = &LVMount;

dinode_t	DIBMap;		/* block allocation map inode */
dinode_t	*diBMap = &DIBMap;

dinode_t	DIIMap;		/* inode allocation map inode */
dinode_t	*diIMap = &DIIMap;

s32	IAGNumber = 0;

/*
 *	file system under defragfs
 */
FS_t	FSMount;	/* file system descriptor */
FS_t	*fsMount = &FSMount;

/* imap xtree sequential read control */
s64	        IMapXtreeLMLeaf; /* imap xtree leftmost leaf */
xtpage_t	IMapXtreeLeaf;   /* imap xtree leaf buffer */
IMapXtree_t	IMapXtree = {&IMapXtreeLeaf, XTENTRYSTART, 1};

/* endian routines */
extern u32 type_jfs;
extern void ujfs_swap_dbmap_t( dbmap_t * );
extern void ujfs_swap_dmapctl_t( dmapctl_t * );
extern void ujfs_swap_dmap_t( dmap_t * );
extern void ujfs_swap_dinomap_t( dinomap_t * );
extern void ujfs_swap_iag_t( iag_t * );
extern void ujfs_swap_xtpage_t( xtpage_t * );

extern int errno;

/*
 *	openLV()
 */
s32 openLV(char	*LVName)
{
	s32	rc = 0;
	s32	pbsize;
	s64	LVSize, hiddenLVSize;
     
    if ( ( rc = Is_Device_Mounted( LVName )) == MSG_JFS_VOLUME_IS_MOUNTED ) {   /* is mounted */
        printf("\n%s is mounted.\n", LVName);
    } else if ( rc == MSG_JFS_VOLUME_IS_MOUNTED_RO ) { /* is mounted read only */
        printf("\n%s is mounted 'read only'.\n", LVName);
    } else if ( rc == MSG_JFS_MNT_LIST_ERROR ) { /* setmntent failed */
        printf("\nCannot detect if %s is mounted.\n", LVName);
        return rc;
    } else if ( rc == MSG_JFS_NOT_JFS ) { /* mounted, but not jfs */
        printf("\nDevice mounted, but not type jfs!\n");
        return rc;
    } else { /* not mounted */
        printf("\n%s is NOT mounted.\n", LVName);
        return -1;
    }

	/* open the device and get its physical block size */
	if (( rc = ujfs_open_device(LVName, &LVHandle, &pbsize, READONLY) ))    {
		if (!Fail_Quietly)
			message_user(MSG_OSO_CANT_OPEN, NULL, 0, STDOUT_CODE,
				     NO_RESPONSE, OSO_MSG);
		return rc;
	}

	LVMount.pbsize = pbsize;
	LVMount.l2pbsize = log2shift(pbsize);
     
	/* get logical volume size */
	if (( rc = ujfs_get_dev_size(LVHandle, &LVSize, &hiddenLVSize)) )    {
		message_user(MSG_OSO_ERR_ACCESSING_DISK, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
		ujfs_close(LVHandle);
		return rc;
	}

	LVMount.LVSize = LVSize >> LVMount.l2pbsize;

	return rc;
}
	

/*
 *	openFS() 
 */ 
s32 openFS(void)
{
	s32	rc;
	struct superblock	*sb;
	buf_t	*bp;

	/*
	 *	validate and retrieve fs parameters from superblock
	 */
	/* try to read the primary superblock */
	rc = bRawRead(LVHandle, (s64)SUPER1_OFF, (s32)PAGESIZE, &bp);
	if (rc != 0)    {
		/* try to read the secondary superblock */
		rc = bRawRead(LVHandle, (s64)SUPER2_OFF, (s32)PAGESIZE, &bp);
		if (rc != 0)       {
			return rc;
		}
	}

	sb = (struct superblock *)bp->b_data;

	/* check magic/version number */
	if (strncmp(sb->s_magic,JFS_MAGIC,(unsigned)strlen(JFS_MAGIC))
	    || (sb->s_version != JFS_VERSION))    {
		message_user(MSG_JFS_BAD_SUPERBLOCK, NULL, 0, STDOUT_CODE, NO_RESPONSE, JFS_MSG);
		return -1;
	}

	if (sb->s_state & FM_DIRTY)    {
		message_user(MSG_JFS_DIRTY, NULL, 0, STDOUT_CODE, NO_RESPONSE, JFS_MSG);
		return -1;
	}

	/* assert(lvMount->pbsize == sb->s_pbsize); */
	/* assert(lvMount->l2pbsize == sb->s_l2pbsize); */

	fsMount->bsize = sb->s_bsize;
	fsMount->l2bsize = sb->s_l2bsize;
	fsMount->l2bfactor = sb->s_l2bfactor;
	fsMount->nbperpage = PAGESIZE >> fsMount->l2bsize;
	fsMount->l2nbperpage = log2shift(fsMount->nbperpage);
	fsMount->FSSize = sb->s_size >> sb->s_l2bfactor;
	fsMount->AGSize = sb->s_agsize;

	bRelease(bp);

	return rc;
}


/*
 *	closeFS()
 */
void closeFS(void)
{
	ujfs_close(LVHandle);
}


/*
 * NAME:	readBMapGCP()
 *
 * FUNCTION:	read the bmap global control page and 
 *		the initial level L2, L1.0 and L0.0 control pages;
 *
 * note: bmap file structure is satic during defragfs()
 * (defragfs() and extendfs() are mutually exclusive operation);
 */
s32 readBMapGCP(BMap_t	*bMap)
{
	s32	rc;
	s64	xaddr;
	s32	xlen;
	dbmap_t	*bcp;

	/* first 4 pages are made contiguous by mkfs() */
	xlen = 4 << fsMount->l2nbperpage;
    rc = xtLookup(diBMap, (s64)0, &xaddr, &xlen, 0);
    if ( rc )
        return rc;

	rc = pRead(fsMount, xaddr, xlen, (void *)bMap);
	if ( rc )    {
		return rc;
	}

	bcp = (dbmap_t *)bMap;

    /* swap if on big endian machine */
    if (type_jfs & JFS_SWAP_BYTES) {
        ujfs_swap_dbmap_t( bcp );
    }

	fsMount->nAG = bcp->dn_numag;

	return 0;
}

		
/*
 * NAME:	readBMapLCP()
 *
 * FUNCTION:  	read level control page of given level for given block number;
 */
s32 readBMapLCP( s64	    bn,
	             s32	    level,
	             dmapctl_t  *lcp)

{
	s32	rc = 0;
	s64	xoff = 0, xaddr;
	s32	xlen;

	switch (level)    {
	case 0:
		xoff = BLKTOL0(bn, fsMount->l2nbperpage);
		break;
	case 1:
		xoff = BLKTOL1(bn, fsMount->l2nbperpage);
		break;
	case 2:
		xoff = fsMount->nbperpage;
		break;
	}

	xlen = fsMount->nbperpage;
	rc = xtLookup(diBMap, xoff, &xaddr, &xlen, 0);		
	if ( rc != 0 )
		return rc;

	rc = pRead(fsMount, xaddr, xlen, (void *)lcp);
	if ( rc )    {
		return rc;
	}

    /* swap if on big endian machine */
    if (type_jfs & JFS_SWAP_BYTES) {
        ujfs_swap_dmapctl_t( lcp );
    }

	return 0;
}


/*
 * NAME:        readBMap()
 *
 * FUNCTION:	Read ndmaps dmap pages starting from the start_dmap.    
 *
 * note: bmap file structure is satic during defragfs()
 * (defragfs() and extendfs() are mutually exclusive operation);
 */
s32 readBMap( s64	  start_dmap,  /* the dmap number to start read */
	          s32	  ndmaps,      /* number of dmap pages read */
	          dmap_t  *dmap)	   /* buffer pointer */
{
    s32	rc;
	s64	xaddr;	
	s64	xoff;
	s32	xlen0, xlen;

	/* convert dmap number to xoffset */
	xaddr = start_dmap << L2BPERDMAP;
    xoff = BLKTODMAP(xaddr, fsMount->l2nbperpage);

	xlen0 = ndmaps << fsMount->l2nbperpage;
	xlen = fsMount->nbperpage;
	for ( ; xlen0 > 0; xlen0 -= xlen) {
        rc = xtLookup(diBMap, xoff, &xaddr, &xlen, 0);
        if ( rc )
            return rc;

		rc = pRead(fsMount, xaddr, xlen, (void *)dmap);
		if ( rc )       {
			return rc;
		}

        /* swap if on big endian machine */
        if (type_jfs & JFS_SWAP_BYTES) {
            ujfs_swap_dmap_t( dmap );
        }

		xoff += xlen;
		dmap = (dmap_t *)((char *)dmap + PAGESIZE);
	}

	return 0;
}


/*
 * NAME:	readIMapGCPSequential()
 *
 * FUNCTION:	read the imap global control page.
 */
s32 readIMapGCPSequential( IMap_t	*iMap,
	                       iag_t	*iag  )
{
	s32	rc;
	xtpage_t	*p;
	s64	xaddr;

	/* read in leftmost leaft xtpage */
	rc = xtLMLeaf(diIMap, &IMapXtreeLeaf); 
	if ( rc != 0 )
		return rc;

	p = &IMapXtreeLeaf;
	IMapXtreeLMLeaf = addressPXD(&p->header.self);

	/* read IMap control page */
	xaddr = addressXAD(&p->xad[XTENTRYSTART]);
	rc = pRead(fsMount, xaddr, fsMount->nbperpage, (void *)iMap);

    /* swap if on big endian machine */
    if (type_jfs & JFS_SWAP_BYTES) {
        ujfs_swap_dinomap_t( &(iMap->ctl) );
    }

	/* init for start of bitmap page read */
	IAGNumber = 0;
	IMapXtree.index = XTENTRYSTART;
	IMapXtree.page = 1;	/* skip for global control page */
	iag->iagnum = -1;

	return rc;
}


/*
 * NAME:	readIMapSequential()
 *
 * FUNCTION:	read one iag page at a time. 
 *
 * state variable:
 *	IMapXtree.leaf - current imap leaf xtpage under scan;
 *	IMapXtree.index - current xad entry index in iMapXtreeLeaf; 
 *	IMapXtree.page - iag number to read within current imap extent; 
 *
 * note: IMap pages may grow (but NOT freed) dynamically; 
 */
s32 readIMapSequential(iag_t	*iag)
{
	s32	rc;
	xtpage_t	*p;
	s64	xaddr;
	s32	xlen, delta;

	p = &IMapXtreeLeaf;

	while(1) {
		/* continue with current leaf xtpage ? */
		if (IMapXtree.index < p->header.nextindex) {
			/* determine iag page extent */
			xaddr = addressXAD(&p->xad[IMapXtree.index]);
			xlen = lengthXAD(&p->xad[IMapXtree.index]);

			if (IMapXtree.page) {
				/* compute offset within current extent */
                delta = IMapXtree.page << fsMount->l2nbperpage;
                xaddr += delta;
                xlen -= delta;
			}

			/* read a iag page */
			rc = pRead(fsMount, xaddr, fsMount->nbperpage,
				       (void *)iag);

            /* swap if on big endian machine */
            if (type_jfs & JFS_SWAP_BYTES) {
                ujfs_swap_iag_t( iag );
            }

			if ( rc != 0 ) {
				return rc;
			}

			/*current extent has more iag */
			if ( xlen > fsMount->nbperpage ) {
				/* continue to read from current extent */
				IMapXtree.page++;
			} else  {
				IMapXtree.index++;
				IMapXtree.page = 0;
			}

			return 0;
		}

		/* read next/right sibling leaf xtpage */
		if (p->header.next != 0) {
			xaddr = p->header.next;

			/* init for next leaf xtpage */
			IMapXtree.index = XTENTRYSTART;
			IMapXtree.page = 0;
		} else {
			/* init for start of bitmap page read */
			IAGNumber = 0;
			IMapXtree.index = XTENTRYSTART;
			IMapXtree.page = 1; /* skip for global control page */

			if (p->header.prev != 0)
				xaddr = IMapXtreeLMLeaf;
 			/* a single leaf xtree */
			else
				continue;
		}

		/* read next xtree leaf page */
		rc = pRead(fsMount, xaddr, fsMount->nbperpage,
			       (void *)&IMapXtreeLeaf);

        /* swap if on big endian machine */
        if (type_jfs & JFS_SWAP_BYTES) {
            ujfs_swap_xtpage_t( &IMapXtreeLeaf );
        }

		if ( rc != 0 ) {
			return rc;
		}
	}
}


/*
 * NAME:        xtLMLeaf()
 *
 * FUNCTION:    read in leftmost leaf page of the xtree
 *              by traversing down leftmost path of xtree;
 */
s32 xtLMLeaf( dinode_t	*dip,	/* disk inode */
              xtpage_t	*pp)	/* pointer to leftmost leaf xtpage */
{
        int       rc;
        xtpage_t  *p;
        s64		  xaddr;

        /* start from root in dinode */
        p = (xtpage_t *)&dip->di_btroot;
        /* is this page leaf ? */
        if (p->header.flag & BT_LEAF) {
		   p->header.next = p->header.prev = 0;
		   memcpy(pp, p, DISIZE);
           return 0;
    	}

        /*
         * traverse down leftmost child node to the leftmost leaf of xtree
         */
        while (TRUE) {
            /* read in the leftmost child page */
	    	xaddr = addressXAD(&p->xad[XTENTRYSTART]);
		    rc = pRead(fsMount, xaddr, fsMount->nbperpage, (void *)pp);
            if (rc) {
                return rc;
	    	}

            /* swap if on big endian machine */
            if (type_jfs & JFS_SWAP_BYTES) {
                ujfs_swap_xtpage_t( pp );
            }

            /* is this page leaf ? */
            if (pp->header.flag & BT_LEAF)
                return 0;
		    else
               	p = pp;
        }
}


/*
 *	 xtree key/entry comparison: extent offset
 *
 * return:
 *	-1: k < start of extent
 *	 0: start_of_extent <= k <= end_of_extent
 *	 1: k > end_of_extent
 */
#define XT_CMP(CMP, K, X, OFFSET64)\
{\
	OFFSET64 = offsetXAD(X);\
	(CMP) = ((K) >= OFFSET64 + lengthXAD(X)) ? 1 :\
	      ((K) < OFFSET64) ? -1 : 0;\
}


/*
 *	xtLookup()
 *
 * function:	search for the xad entry covering specified offset.
 *
 * parameters:
 *	ip	    - file object;
 *	xoff	- extent offset;
 *	cmpp	- comparison result:
 *	btstack	- traverse stack;
 *	flag	- search process flag (XT_INSERT);
 *
 * returns:
 *	btstack contains (bn, index) of search path traversed to the entry.
 *	*cmpp is set to result of comparison with the entry returned.
 *	the page containing the entry is pinned at exit.
 */
s32 xtLookup( dinode_t	*dip,
	          s64		xoff,	/* offset of extent */
	          s64		*xaddr,
	          s32		*xlen,
	          u32		flag)
{
	s32		rc = 0;
	buf_t		*bp = NULL;	/* page buffer */
	xtpage_t	*p;	/* page */
	xad_t		*xad;
	s64		t64;
	s32		cmp = 1, base, index, lim;
	s32		t32;

	/*
	 *	search down tree from root:
	 *
	 * between two consecutive entries of <Ki, Pi> and <Kj, Pj> of
	 * internal page, child page Pi contains entry with k, Ki <= K < Kj.  
	 *
 	 * if entry with search key K is not found
 	 * internal page search find the entry with largest key Ki 
	 * less than K which point to the child page to search;
 	 * leaf page search find the entry with smallest key Kj 
	 * greater than K so that the returned index is the position of
	 * the entry to be shifted right for insertion of new entry.
	 * for empty tree, search key is greater than any key of the tree.
	 */

	/* start with inline root */
	p = (xtpage_t *)&dip->di_btroot;

xtSearchPage:
	lim = p->header.nextindex - XTENTRYSTART; 

	/*
	 * binary search with search key K on the current page
	 */
	for (base = XTENTRYSTART; lim; lim >>= 1) {
		index = base + (lim >> 1);

		XT_CMP(cmp, xoff, &p->xad[index], t64);
		if (cmp == 0) {
			/*
			 *	search hit
			 */
			/* search hit - leaf page:
			 * return the entry found
			 */
			if (p->header.flag & BT_LEAF) {
				/* save search result */
				xad = &p->xad[index];
				t32 = xoff - offsetXAD(xad);
				*xaddr = addressXAD(xad) + t32; 
				*xlen = MIN(*xlen, lengthXAD(xad) - t32);

				rc = cmp;
				goto out;
			}

			/* search hit - internal page:
			 * descend/search its child page 
			 */
			goto descend;
		}

		if (cmp > 0) {
			base = index + 1;
			--lim;
		}
	}

	/*
	 *	search miss
	 *
	 * base is the smallest index with key (Kj) greater than 
	 * search key (K) and may be zero or maxentry index.
	 */
	/*
	 * search miss - leaf page: 
	 *
	 * return location of entry (base) where new entry with
	 * search key K is to be inserted.
	 */
	if (p->header.flag & BT_LEAF) {
		rc = cmp;
		goto out;
	}

	/*
	 * search miss - non-leaf page:
	 *
	 * if base is non-zero, decrement base by one to get the parent
	 * entry of the child page to search.
	 */
	index = base ? base - 1 : base;

	/* 
	 * go down to child page
	 */
descend:
	/* get the child page block number */
	t64 = addressXAD(&p->xad[index]);

	/* release parent page */
	if (bp)
		bRelease(bp);

	/* read child page */
	bRead(fsMount, t64, fsMount->nbperpage, &bp);

	p = (xtpage_t *)bp->b_data;
	goto xtSearchPage;

out:
	/* release current page */
	if (bp)
		bRelease(bp);

	return rc;
}


/*
 * NAME:	fscntl()
 *
 * FUNCTION:	call fs control service;
 */
s32 fscntl(
	u32	cmd,
	void	*pList,		/* parameter list */
	ULONG	*pListLen,	/* length of pList */
	void	*pData,		/* data area */
	ULONG	*pDataLen)	/* length of pData */
{
	s32	rc = 0;

	return rc;
}


/*
 *	pRead()
 *
 * read into specified buffer;
 */
s32 pRead( FS_t	 *fsMount,
	       s64	 xaddr,	   /* in bsize */
	       s32	 xlen,	   /* in bsize */
	       void	 *p)	   /* buffer */
{
	s32	rc;
	s64	off;	/* offset in byte */
	s32	len;	/* length in byte */

	/* read in from disk */
	off = xaddr << fsMount->l2bsize;
	len = xlen << fsMount->l2bsize;
	if ( (rc = ujfs_rw_diskblocks(LVHandle, off, len, p, GET)) ) {
		message_user(MSG_OSO_READ_ERROR, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
	}

	return rc;
}


/*
 *	buffer pool
 *	===========
 */
s32	NBUFFER = 8;
s32	nBuffer = 0;
buf_t	*bCachelist = NULL;

static void bAllocate(void)
{
	buf_t	*bp;
	char	*buffer;
	s32	i;

	/* allocate buffer header */
	if ((bp = malloc(NBUFFER * sizeof(buf_t))) == NULL) {
		message_user(MSG_OSO_NOT_ENOUGH_MEMORY, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
		exit (ERROR_DISK_FULL);
	}

	/* allocate buffer pages */
	if ((buffer = malloc(NBUFFER * PAGESIZE)) == NULL) {
		message_user(MSG_OSO_NOT_ENOUGH_MEMORY, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
		exit (ERROR_DISK_FULL);
	}

	/* insert buffer headers in lru/freelist */
	for (i = 0; i < NBUFFER; i++, bp++, buffer += PAGESIZE) {
		bp->b_data = buffer;
		bp->b_next = bCachelist;
		bCachelist = bp;
	}

	nBuffer += NBUFFER;
}

/*
 *	bRead()
 *
 * aloocate, read and return a buffer;
 */
s32 bRead( FS_t	  *fsMount,
	       s64	  xaddr,	/* in bsize */
	       s32	  xlen,	    /* in bsize */
	       buf_t  **bpp)
{
	s32	rc;
	s64	off;	/* offset in byte */
	s32	len;	/* length in byte */
	buf_t	*bp;

	/* allocate buffer */
	if (bCachelist == NULL)
		bAllocate();
	bp = bCachelist;
	bCachelist = bp->b_next;

	/* read in from disk */
	off = xaddr << fsMount->l2bsize;
	len = xlen << fsMount->l2bsize;
	rc = ujfs_rw_diskblocks(LVHandle, off, len, (void *)bp->b_data, GET);
	if (rc == 0)
		*bpp = bp;
    else {
		message_user(MSG_OSO_READ_ERROR, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
		bRelease(bp);
	}

	return rc;
}

s32 bRawRead( u32	 LVHandle,
	          s64	 off,	/* in byte */
	          s32	 len,	/* in byte */
	          buf_t  **bpp)
{
	s32	rc;
	buf_t	*bp;

	/* allocate buffer */
	if (bCachelist == NULL)
		bAllocate();
	bp = bCachelist;
	bCachelist = bp->b_next;

	/* read in from disk */
	rc = ujfs_rw_diskblocks(LVHandle, off, len, (void *)bp->b_data, GET);
	if (rc == 0)
		*bpp = bp;
    else {
		message_user(MSG_OSO_READ_ERROR, NULL, 0, STDOUT_CODE, NO_RESPONSE, OSO_MSG);
		bRelease(bp);
	}

	return rc;
}

void bRelease(buf_t	*bp)
{
	/* release buffer */
	bp->b_next = bCachelist;
	bCachelist = bp;
}
 
/*--------------------------------------------------------------------
 * NAME: Is_Device_Mounted
 *
 * FUNCTION: Determine if the device specified is mounted.
 *
 * PRE CONDITIONS: Device_Name must be the name of the device.
 *
 * POST CONDITIONS:
 *
 * PARAMETERS: Device_Name - The name of the device.
 *
 * RETURNS:
 *      0 if the device is not mounted.
 *    > 0 if the device is mounted or an error occurs.
 *
 */
int Is_Device_Mounted(char * Device_Name)
{

  FILE *          Mount_Records;  /* Pointer for system's mount records */
  struct mntent * Mount_Data;     /* Holds data for entry in mount list */
  int             fd, Is_Mounted = 0; /* Holds return code.             */
  char *          opt;            

  /* Prepare the mnt stream for access. */
  Mount_Records = setmntent(MOUNTED,"r");
  if ( Mount_Records == NULL ) {
    /* We are unable to access the list of mounted devices!   */

    /* Check (the hard way) if we're mounted 'ro' (read only) */
    /* If so, we're probably booting and there's a problem    */
    /* accessing /etc/mtab.  In that case, the log needs to   */
    /* to be replayed.                                        */
/*  fd = open(MOUNTED, O_RDWR);
    if (fd < 0) {
      if (errno == EROFS)
        / root is mounted ro (read only), let caller know / 
        return MSG_JFS_VOLUME_IS_MOUNTED_RO;
    } else {
      close(fd); */
      return MSG_JFS_MNT_LIST_ERROR;
/*  } */
  }

  do {
    /* Get an entry from the list of mounted devices. */
    Mount_Data  = getmntent(Mount_Records);

    /* Did we get the data or are we at the end of the list? */
    if ( Mount_Data != NULL ) {
      /* Does the current mount record describe the device we are looking for? */
      if ( ( strcmp(Device_Name, Mount_Data->mnt_fsname) == 0 ) ||
           ( strcmp(Device_Name, Mount_Data->mnt_dir)    == 0 ) )  {
        /* We have a match!  The device is mounted! See if it's jfs */
        if ( strcmp(Mount_Data->mnt_type, "jfs") == 0 ) {
          /* is mounted, is type jfs */
          Is_Mounted = MSG_JFS_VOLUME_IS_MOUNTED;

          /* see if we're mounted 'ro' (read only) */

          /* if we are booting on a jfs device, getmntent will   */
          /* probably return 'rw' even though we're mounted 'ro' */
          /* so let's check for that situation                   */
          if (!strcmp(Mount_Data->mnt_dir, "/")) {
            fd = open(MOUNTED, O_RDWR);
            if (fd < 0) {
              if (errno == EROFS)
                /* root is mounted ro (read only), let caller know */ 
                Is_Mounted = MSG_JFS_VOLUME_IS_MOUNTED_RO;
            } else
              close(fd);
          } else {
            opt = Mount_Data->mnt_opts;
            while ( (*opt != '\0') && (Is_Mounted == MSG_JFS_VOLUME_IS_MOUNTED) ) {
              if ( *opt == 'r' && *(opt+1) == 'o'  &&
                  (*(opt+2) == ',' || *(opt+2) == '\0') )   { 
                /* option is ro (read only), let caller know */ 
                Is_Mounted = MSG_JFS_VOLUME_IS_MOUNTED_RO;
              } else {
                /* move pointer past comma or to 'end of string' */
                while (*opt != ',' && *opt != '\0') opt++;
                if (*opt == ',') opt++;
              }
            }
          }
        } else {
          /* is mounted, is not type jfs */
          Is_Mounted = MSG_JFS_NOT_JFS;
        }
      }
    }

  } while ( ( Mount_Data != NULL ) && ( ! Is_Mounted ) );

  /* We finished processing all of the mount records ( or there was an error which we will ignore ). */

  /* Close the stream. */
  endmntent(Mount_Records);

  /* Return the appropriate value. */
  return Is_Mounted;

}

/*--------------------------------------------------------------------
 * NAME: Is_Device_Type_JFS
 *
 * FUNCTION: Determine if the device specified is of type JFS in /etc/fstab
 *
 * PRE CONDITIONS: Device_Name must be the name of the device.
 *
 * POST CONDITIONS:
 *
 * PARAMETERS: Device_Name - The name of the device.
 *
 * RETURNS:
 *      0 if the device is type JFS in /etc/fstab
 *    > 0 if the device is not type JFS or an error occurs.
 *
 */
int Is_Device_Type_JFS(char * Device_Name)
{

  FILE *          Mount_Table;   /* Pointer for system's mount table      */
  struct mntent * Mount_Data;    /* Holds data for entry in mount list    */
  int             Is_JFS = 0;    /* Holds return code.                    */
  int             In_Table = 0;  /* Signifies if device is in mount table */

  /* Prepare the mnt stream for access. */
  Mount_Table = setmntent(_PATH_MNTTAB,"r");
  if ( Mount_Table == NULL ) {
    /* We are unable to access the list of devices!   */
    return MSG_JFS_MNT_LIST_ERROR;
  }

  do {
    /* Get an entry from the list of devices. */
    Mount_Data  = getmntent(Mount_Table);

    /* Did we get the data or are we at the end of the list? */
    if ( Mount_Data != NULL ) {
      /* Does the current mount record describe the device we are looking for? */
      if ( ( strcmp(Device_Name, Mount_Data->mnt_fsname) == 0 ) ||
           ( strcmp(Device_Name, Mount_Data->mnt_dir)    == 0 ) )  {
        /* We have a match! */
        In_Table = 1;
        /* See if it's jfs */
        if ( strcmp(Mount_Data->mnt_type, "jfs") != 0 ) {
          /* is not type jfs */
          Is_JFS = MSG_JFS_NOT_JFS;
        }
      }
    }
  } while ( ( Mount_Data != NULL ) && ( ! In_Table ) );
  /* We finished processing all of the mount records ( or there was an error which we will ignore ). */

  /* Close the stream. */
  endmntent(Mount_Table);

  if ( ! In_Table ) 
    Is_JFS = MSG_JFS_DEV_NOT_IN_TABLE;

  /* Return the appropriate value. */
  return Is_JFS;

}
